
/*
 * This program source code file is part of SystemK, a library in the KTag project.
 *
 * 🛡️ <https://ktag.clubk.club> 🃞
 *
 * Copyright © 2016-2025 Joseph P. Kearney and the KTag developers.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
 * details.
 *
 * There should be a copy of the GNU Affero General Public License in the LICENSE
 * file in the root of this repository.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdbool.h>
#include <stdio.h>

#include "SystemK.h"

#define MILES_TAG_RED_TEAM 0b00
#define MILES_TAG_BLUE_TEAM 0b01
#define MILES_TAG_YELLOW_TEAM 0b10
#define MILES_TAG_GREEN_TEAM 0b11

#define MILES_TAG_II_HEADER_PULSE_DURATION_IN_us 2400
#define MILES_TAG_II_SPACE_DURATION_IN_us 600
#define MILES_TAG_II_MARK_ZERO_DURATION_IN_us 600
#define MILES_TAG_II_MARK_ONE_DURATION_IN_us 1200
#define MILES_TAG_II_TOLERANCE_IN_us 300

#define MIN_SPACE_IN_us (MILES_TAG_II_SPACE_DURATION_IN_us - MILES_TAG_II_TOLERANCE_IN_us)
#define MAX_SPACE_IN_us (MILES_TAG_II_SPACE_DURATION_IN_us + MILES_TAG_II_TOLERANCE_IN_us)
#define MIN_MARK_ZERO_IN_us (MILES_TAG_II_MARK_ZERO_DURATION_IN_us - MILES_TAG_II_TOLERANCE_IN_us)
#define MAX_MARK_ZERO_IN_us (MILES_TAG_II_MARK_ZERO_DURATION_IN_us + MILES_TAG_II_TOLERANCE_IN_us)
#define MIN_MARK_ONE_IN_us (MILES_TAG_II_MARK_ONE_DURATION_IN_us - MILES_TAG_II_TOLERANCE_IN_us)
#define MAX_MARK_ONE_IN_us (MILES_TAG_II_MARK_ONE_DURATION_IN_us + MILES_TAG_II_TOLERANCE_IN_us)

static const TimedBit_T MILES_TAG_HEADER = {.symbol = MARK, .duration = MILES_TAG_II_HEADER_PULSE_DURATION_IN_us};
static const TimedBit_T MILES_TAG_SPACE = {.symbol = SPACE, .duration = MILES_TAG_II_SPACE_DURATION_IN_us};
static const TimedBit_T MILES_TAG_ONE = {.symbol = MARK, .duration = MILES_TAG_II_MARK_ONE_DURATION_IN_us};
static const TimedBit_T MILES_TAG_ZERO = {.symbol = MARK, .duration = MILES_TAG_II_MARK_ZERO_DURATION_IN_us};
static const TimedBit_T MILES_TAG_END = {.symbol = SPACE, .duration = LAST_PULSE};

static TimedPulseTrain_T Tag_Send_Buffer;
static DecodedPacket_T Miles_Tag_II_Tag_Received_Buffers[3] =
    {
        {.Tag.type = DECODED_PACKET_TYPE_BUFFER_FREE},
        {.Tag.type = DECODED_PACKET_TYPE_BUFFER_FREE},
        {.Tag.type = DECODED_PACKET_TYPE_BUFFER_FREE}};

static inline DecodedPacket_T *Get_Tag_Packet_Buffer(void)
{
    if (Miles_Tag_II_Tag_Received_Buffers[0].Tag.type == DECODED_PACKET_TYPE_BUFFER_FREE)
    {
        return &Miles_Tag_II_Tag_Received_Buffers[0];
    }
    else if (Miles_Tag_II_Tag_Received_Buffers[1].Tag.type == DECODED_PACKET_TYPE_BUFFER_FREE)
    {
        return &Miles_Tag_II_Tag_Received_Buffers[1];
    }
    else
    {
        // Just use it.
        return &Miles_Tag_II_Tag_Received_Buffers[2];
    }
}

static const uint16_t Miles_Tag_II_Damage[] = {1, 2, 4, 5, 7, 10, 15, 17, 20, 25, 30, 35, 40, 50, 75, 100};
static const color_t Miles_Tag_II_Team_Colors[] = {COLOR_RED, COLOR_BLUE, COLOR_YELLOW, COLOR_GREEN};

static inline uint8_t Encode_Damage(uint16_t damage)
{
    uint8_t encoded_damage;

    for (encoded_damage = 0; encoded_damage < 16; encoded_damage++)
    {
        if (damage <= Miles_Tag_II_Damage[encoded_damage])
        {
            break;
        }
    }
    return encoded_damage;
}

// Use the 14 least-significant bits.
static inline void PackPulseTrain(TimedPulseTrain_T *pulsetrain, uint16_t data)
{
    pulsetrain->bitstream[0] = MILES_TAG_HEADER;
    pulsetrain->bitstream[1] = MILES_TAG_SPACE;

    for (uint8_t n = 0; n < 14; n++)
    {
        if ((data & 0x01) == 0x01)
        {
            pulsetrain->bitstream[2 + (n * 2)] = MILES_TAG_ONE;
        }
        else
        {
            pulsetrain->bitstream[2 + (n * 2)] = MILES_TAG_ZERO;
        }
        pulsetrain->bitstream[2 + (n * 2) + 1] = MILES_TAG_SPACE;
        data = data >> 1;
    }

    pulsetrain->bitstream[30] = MILES_TAG_END;
}

TimedPulseTrain_T *MILES_TAG_II_EncodePacket(TagPacket_T *packet)
{
    uint16_t packed_data = Encode_Damage(packet->damage);
    packed_data |= (packet->team_ID << 4);
    packed_data |= (packet->player_ID << 6);

    PackPulseTrain(&Tag_Send_Buffer, packed_data);

    return &Tag_Send_Buffer;
}

DecodedPacket_T *MILES_TAG_II_MaybeDecodePacket(TimedPulseTrain_T *packet)
{
    // Some implementations (ESP32) do not report odd numbers of pulses.
    if ((packet->count != 29) && (packet->count != 30))
    {
        // Fail fast!
        return NULL;
    }
    else
    {
        // Convert pulse durations to bits.
        //  packet->bitstream[0] is the header--ignore it.
        //  packet->bitstream[1] is the space after the header--ignore it also.
        uint16_t data = 0;
        for (uint8_t n = 0; n < 14; n++)
        {
            // Even pulses are marks; odd pulses are spaces.
            // We only check the marks against the tolerances, to save duration.
            if ((packet->bitstream[2 + (n * 2)].duration > MIN_MARK_ONE_IN_us) && (packet->bitstream[2 + (n * 2)].duration < MAX_MARK_ONE_IN_us))
            {
                // One
                data |= ((uint16_t)1 << n);
            }
            else if ((packet->bitstream[2 + (n * 2)].duration > MIN_MARK_ZERO_IN_us) && (packet->bitstream[2 + (n * 2)].duration < MAX_MARK_ZERO_IN_us))
            {
                // Zero--nothing to do.
            }
            else
            {
                // This mark is neither a zero or a one; abort;
                return NULL;
            }
        }

        DecodedPacket_T *Tag_Rx_Buffer = Get_Tag_Packet_Buffer();
        Tag_Rx_Buffer->Tag.type = DECODED_PACKET_TYPE_TAG_RECEIVED;
        Tag_Rx_Buffer->Tag.protocol = MILES_TAG_II_PROTOCOL;
        Tag_Rx_Buffer->Tag.player_ID = 0x007F & (data >> 6);
        Tag_Rx_Buffer->Tag.team_ID = 0x0003 & (data >> 4);
        Tag_Rx_Buffer->Tag.damage = Miles_Tag_II_Damage[0x000F & data];
        Tag_Rx_Buffer->Tag.color = Miles_Tag_II_Team_Colors[Tag_Rx_Buffer->Tag.team_ID];

        return Tag_Rx_Buffer;
    }
}

color_t MILES_TAG_II_GetTeamColor(uint8_t team_ID)
{
    color_t result = COLOR_RED;

    if (team_ID == MILES_TAG_RED_TEAM)
    {
        // Nothing to do.
    }
    else if (team_ID == MILES_TAG_BLUE_TEAM)
    {
        result = COLOR_BLUE;
    }
    else if (team_ID == MILES_TAG_YELLOW_TEAM)
    {
        result = COLOR_YELLOW;
    }
    else if (team_ID == MILES_TAG_GREEN_TEAM)
    {
        result = COLOR_GREEN;
    }

    return result;
}
